/*
* DCF77 decoding
* (c) 27.6.2015 Ingo D. Rullhusen
*/


const unsigned long dcf_bit_len = 28800;   // 150ms 192k*0.15
const unsigned long dcf_min_len = 288000;  // 192k*1.5

unsigned long dcf_sec_cnt;

int next_dcf_bit;
char dcf_bits[60];


typedef struct
{
   char ERROR;
   char M;
   unsigned int bits_1_14;
   char R, A1, Z1, Z2, A2, S;
   unsigned char Min, Hr, EST;
   unsigned char Day, WDay, Month, Yr;
} DCF_DATA;

DCF_DATA dcf_data;

void reset_dcf()
{
   dcf_sec_cnt = 0;
   next_dcf_bit = 0;
}

void setup_dcf_info()
{
   int i, mask;

   dcf_data.ERROR = 0;
   dcf_data.M = dcf_bits[0];
   if ( dcf_data.M > 0 ) { dcf_data.ERROR += 1; }

   dcf_data.bits_1_14 = 0;
   mask = 1;
   for ( i=1; i<15; ++i )
   {
      if ( dcf_bits[i]> 0 )
      {
         dcf_data.bits_1_14 |= mask;
      }
      mask<<=1;
   }

   dcf_data.R = dcf_bits[15];
   dcf_data.A1= dcf_bits[16];
   dcf_data.Z1= dcf_bits[17];
   dcf_data.Z2= dcf_bits[18];
   dcf_data.A2= dcf_bits[19];
   dcf_data.S = dcf_bits[20];

   if      ( (dcf_data.Z1==0) && (dcf_data.Z2==1) )
   { dcf_data.EST = 0; }
   else if ( (dcf_data.Z1==1) && (dcf_data.Z2==0) )
   { dcf_data.EST = 1; }
   else
   { dcf_data.ERROR += 1; }

   if ( dcf_data.S != 1 ) { dcf_data.ERROR += 1; }

   dcf_data.Min = (dcf_bits[21]>0 ? 0x1 : 0x0 ) |
                  (dcf_bits[22]>0 ? 0x2 : 0x0 ) |
                  (dcf_bits[23]>0 ? 0x4 : 0x0 ) |
                  (dcf_bits[24]>0 ? 0x8 : 0x0 );

   dcf_data.Min += 10*( (dcf_bits[25]>0 ? 0x1 : 0x0 ) |
                        (dcf_bits[26]>0 ? 0x2 : 0x0 ) |
                        (dcf_bits[27]>0 ? 0x4 : 0x0 ) );

   mask = 0;  // P1
   for ( i=21; i<29; ++i )
   { mask += dcf_bits[i]; }
   if ( mask % 2 ) { dcf_data.ERROR += 1; }

   dcf_data.Hr = (dcf_bits[29]>0 ? 0x1 : 0x0 ) |
                 (dcf_bits[30]>0 ? 0x2 : 0x0 ) |
                 (dcf_bits[31]>0 ? 0x4 : 0x0 ) |
                 (dcf_bits[32]>0 ? 0x8 : 0x0 );

   dcf_data.Hr += 10*( (dcf_bits[33]>0 ? 0x1 : 0x0 ) |
                       (dcf_bits[34]>0 ? 0x2 : 0x0 ) );

   mask = 0;  // P2
   for ( i=29; i<36; ++i )
   { mask += dcf_bits[i]; }
   if ( mask % 2 ) { dcf_data.ERROR += 1; }


   dcf_data.Day = (dcf_bits[36]>0 ? 0x1 : 0x0 ) |
                  (dcf_bits[37]>0 ? 0x2 : 0x0 ) |
                  (dcf_bits[38]>0 ? 0x4 : 0x0 ) |
                  (dcf_bits[39]>0 ? 0x8 : 0x0 );

   dcf_data.Day += 10*( (dcf_bits[40]>0 ? 0x1 : 0x0 ) |
                        (dcf_bits[41]>0 ? 0x2 : 0x0 ) );
   if ( dcf_data.Day<1 || dcf_data.Day>31 ) { dcf_data.ERROR += 1; }

   dcf_data.WDay = (dcf_bits[42]>0 ? 0x1 : 0x0 ) |
                   (dcf_bits[43]>0 ? 0x2 : 0x0 ) |
                   (dcf_bits[44]>0 ? 0x4 : 0x0 );
   if ( dcf_data.WDay<1 || dcf_data.WDay>7 ) { dcf_data.ERROR += 1; }

   dcf_data.Month = (dcf_bits[45]>0 ? 0x1 : 0x0 ) |
                    (dcf_bits[46]>0 ? 0x2 : 0x0 ) |
                    (dcf_bits[47]>0 ? 0x4 : 0x0 ) |
                    (dcf_bits[48]>0 ? 0x8 : 0x0 );

   dcf_data.Month += (dcf_bits[49]>0 ? 10 : 0x0 );
   if ( dcf_data.Month<1 || dcf_data.Month>12 ) { dcf_data.ERROR += 1; }

   dcf_data.Yr = (dcf_bits[50]>0 ? 0x1 : 0x0 ) |
                 (dcf_bits[51]>0 ? 0x2 : 0x0 ) |
                 (dcf_bits[52]>0 ? 0x4 : 0x0 ) |
                 (dcf_bits[53]>0 ? 0x8 : 0x0 );

   dcf_data.Yr += 10*( (dcf_bits[54]>0 ? 0x1 : 0x0 ) |
                       (dcf_bits[55]>0 ? 0x2 : 0x0 ) |
                       (dcf_bits[56]>0 ? 0x4 : 0x0 ) |
                       (dcf_bits[57]>0 ? 0x8 : 0x0 ) );

   mask = 0;  // P3
   for ( i=36; i<59; ++i )
   { mask += dcf_bits[i]; }
   if ( mask % 2 ) { dcf_data.ERROR += 1; }

}

const char* sWday( int d )
{
   switch(d)
   {
      case 1: { return( "Mo." ); }
      case 2: { return( "Di." ); }
      case 3: { return( "Mi." ); }
      case 4: { return( "Do." ); }
      case 5: { return( "Fr." ); }
      case 6: { return( "Sa." ); }
      case 7: { return( "So." ); }
   }
   return("***");
}

void print_dcf_info()
{
   setup_dcf_info();

   printf("%s 0x%0.4x %.2i:%.2i %s %s %0.2i.%0.2i.%0.2i\n",
          (dcf_data.ERROR>0 ? "*****" : "DCF77" ),
          dcf_data.bits_1_14,
          dcf_data.Hr,dcf_data.Min, (dcf_data.EST>0 ? "MESZ" : "MEZ " ),
          sWday( dcf_data.WDay ),
          dcf_data.Day,
          dcf_data.Month,
          dcf_data.Yr
         );

   fflush(stdout);
}


int to0_dcf( long smplno )
{
   int i;
   int trigger = 0;

   dcf_sec_cnt += smplno;
   if ( dcf_sec_cnt > dcf_min_len )
   {
      print_dcf_info();
      next_dcf_bit = 0;
      for ( i=0; i<60; ++i ) { dcf_bits[i] = 0; }
      trigger = 1;
   }
   dcf_sec_cnt = 0;

   return(trigger);
}

void to1_dcf( long smplno )
{
   dcf_sec_cnt += smplno;
   if ( dcf_sec_cnt > dcf_bit_len )
   {
      dcf_bits[ next_dcf_bit ] = 1;
   }
   else
   {
      dcf_bits[ next_dcf_bit ] = 0;
   }
   ++next_dcf_bit;
   if ( next_dcf_bit >= 60 ) { next_dcf_bit = 0; }
}

void nochange_dcf( long smplno )
{
   dcf_sec_cnt += smplno;
}
